#!/usr/bin/perl -w
#
# Manage changes between existing data sets and new
# date sets, generating a report on them.
#
# We compare the values stored in a hashref and return
# an arrayref containing a description of those changes.
#
# Example usage:
#
# my $changes = Changes::compare(
#       undef,   # any $changes hashref to update
#       'Item',  # the name of this item
#       $oldref, $newref, # refs to compare
#     );
# 
# $changes is an arrayref containing a hashref of the
# changes which were detected:
#
#     'type'   => 'add', 'remove', 'change'
#     'key'    => Key affected
#     'title'  => Description of the change
#     'desc'   => Readable description of the change
#
# Advanced users may use a dedicated description
# function to describe the changes, eg:
#
# my $changes = Changes::compare(
#       undef,   # any $changes hashref to update
#       'Item',  # the name of this item
#       $oldref, $newref, # refs to compare
#       sub {
#           my ($key, $ref) = @_;
#           return "Product: $key\n" .
#                  "Price:   $ref->{'price'}";
#       }
#     );
#
# The state of the hashref can be stored so that
# it can be restored in future executions. For
# example:
#
# my $changes = Changes::compareStored(
#       undef,   # any $changes hashref to update
#       'Item',  # the name of this item
#       'items', # leafname of the item data to preserve
#       $newref, # new ref data to compare
#     );
#
# To display the changes, an example function is
# provided which will return the differences.
#
# print Changes::toText($changes);
#

package Changes;

use Storable;

our $storeDir = '/var/jfcache/changes';


##
# Produce a comparison of an old dataset (from disc) and new dataset
#
# @param[in] $changes   List of changes to add to, or undef for empty set
# @param[in] $type      The type of change for the description
# @param[in] $leafname  Leafname of the dataset on disc
# @param[in] $new       A hashref of values to check (the new ones)
# @param[in] $func      A function to produce a description of the values
#
# @return $changes arrayref of changes, disc file updated
sub compareStored
{
    my ($changes, $type, $leafname, $new, $func) = @_;
    my $store = "$storeDir/$leafname";
    my $old = {};
    
    if (-f $store)
    {
        $old = retrieve $store;
    }
    
    $changes = compare($changes, $type, $old, $new, $func);
    
    # Update the state on disc
    mkdir $storeDir if (!-d $storeDir);
    store $new, $store;
    
    return $changes;
}

##
# Compare two hashrefs for differences
#
# @param[in] $changes   List of changes to add to, or undef for empty set
# @param[in] $type      The type of change for the description
# @param[in] $old       A hashref of values to check (the old ones)
# @param[in] $new       A hashref of values to check (the new ones)
# @param[in] $func      A function to produce a description of the values
#
# @return $changes arrayref of changes
sub compare
{
    my ($changes, $type, $old, $new, $func) = @_;
    my $key;
    $changes = [] if (!defined $changes);

    $func = \&defaultcompare if (!defined $func);;
    
    # Check for new/changed items
    for $key (sort keys %$new)
    {
        if (!defined $old->{$key})
        {
            added($changes, $type, $key,
                  & $func ( $new->{$key}, $key ) );
        }
        else
        {
            my $newDef = & $func ( $new->{$key}, $key );
            my $oldDef = & $func ( $old->{$key}, $key );
            #print "$key:\n",
            #      map({ "OLD: $_\n" } split /\n/, $oldDef),
            #      map({ "NEW: $_\n" } split /\n/, $newDef);
            if ($newDef ne $oldDef)
            {
                changed( $changes, $type, $key,
                         $oldDef, $newDef );
            }
        }
    }

    # Check for old items
    for $key (sort keys %$old)
    {
        if (!defined $new->{$key})
        {
            deleted($changes, $type, $key,
                    & $func ( $old->{$key}, $key ) );
        }
    }
    
    return $changes;
}


##
# A default comparison function which takes a value
# and produces a simple table of the key values.
# Can handle:
#
#   strings
#   hashrefs of key-string pairs
#   arrayrefs of strings
#
# @param[in] $ref   The value to produce description of
# @param[in] $name  The name of the value we're comparing
sub defaultcompare
{
    my ($ref, $name) = @_;
    my @list;
    if (ref $ref eq 'HASH')
    {
        my $longest = 3;
        for my $key (keys %$ref)
        {
            $longest = length $key if (length $key > $longest);
        }
        push @list, sprintf("%-*s %s", $longest+1, "Key:", $name);
        push @list, map {
                        sprintf("%-*s %s", $longest+1, "$_:", $ref->{$_})
                    } sort keys %$ref;
    }
    elsif (ref $ref eq 'ARRAY')
    {
        my $longest = 0;
        my $total;
        my $comma = 0;
        for my $val (@$ref)
        {
            $longest = length $val if (length $val > $longest);
            $total += length $val;
            $comma = 1 if ($val =~ /,/);
        }
        if ($longest > 20 ||
            $total > 200 ||
            $comma)
        {
            # One entry per line if there's long, or lots of entries
            push @list, @$ref;
        }
        else
        {
            # Comma separated if there's few items
            push @list, join(", ",  @$ref);
        }
    }
    elsif (ref $ref eq '')
    {
        return $ref;
    }
    return join "\n", @list;
}


##
# Update changeset with an 'add' change.
#
# @param[in] $changes   The changes to update
# @param[in] $type      The type of change to prefix the title with
# @param[in] $key       The key name to describe
# @param[in] $desc      The textual description
sub added
{
    my ($changes, $type, $key, $desc) = @_;
    $desc =~ s/\n+$//;
    push @$changes, {
        'type'  => 'add',
        'key'   => $key,
        'title' => "$type '$key' added",
        'desc'  => $desc
    };
}



##
# Update changeset with a 'delete' change.
#
# @param[in] $changes   The changes to update
# @param[in] $type      The type of change to prefix the title with
# @param[in] $key       The key name to describe
# @param[in] $desc      The textual description
sub deleted
{
    my ($changes, $type, $key, $desc) = @_;
    $desc =~ s/\n+$//;
    push @$changes, {
        'type'  => 'remove',
        'key'   => $key,
        'title' => "$type '$key' removed",
        'desc'  => $desc
    };
}



##
# Update changeset with a 'change' change.
#
# @param[in] $changes   The changes to update
# @param[in] $type      The type of change to prefix the title with
# @param[in] $key       The key name to describe
# @param[in] $olddesc   The old-textual description
# @param[in] $newdesc   The new-textual description
sub changed
{
    my ($changes, $type, $key, $olddesc, $newdesc) = @_;
    $newdesc =~ s/\n+$//;
    push @$changes, {
        'type'  => 'change',
        'key'   => $key,
        'title' => "$type '$key' changed",
        'desc'  => "Now:\n" .
                   join("", map { "  $_\n" } split /\n/, $newdesc).
                   "Was:\n" .
                   join("", map { "  $_\n" } split /\n/, $olddesc)
    };
}


##
# Convert a change list to a string.
#
# @param[in] $changes  The changes to convert
#
# @return A printable text description of the changes
sub toText
{
    my ($changes) = @_;
    my $text = "";
    for my $change (@$changes)
    {
        my $title = $change->{'title'};
        my $desc = $change->{'desc'};
        $text .= "$title\n";
        $text .= '-' x length($title),
        $text .= "\n";
        $text .= "$desc\n";
        $text .= "\n\n\n";
    }
    return $text;
}


1;
