#!/usr/bin/perl -w
#
# Generate an RSS page containing the details of what's changed
# recently by applying a change to the existing set.
#
# See Changes.pm for the generation of change lists.

package ChangesRSS;

use Changes;
use Storable;
use Digest::MD5;

# 0 = escape HTML
# 1 = use CDATA
my $type = 1;


##
# Generate an RSS file, for a given change set.
#
# @param[in,out] $changes   Changes generated by Changes.pm
# @param[in] $filename      Filename to write to
# @param[in] $name          Dataset name (leafname for storing old results)
# @param[in] $params        Parameters hashref for RSS generation see
#                           generate().
sub generateFile
{
    my ($changes, $filename, $name, $params) = @_;
    my $rss = generate($changes, $name, $params);
    open(my $fh, '>', $filename) || die "Failed to write to $filename: $!\n";
    print $fh $rss;
    close($fh);
}


##
# Generate RSS content, for a given change set.
#
# @param[in,out] $changes   Changes generated by Changes.pm
# @param[in] $name          Dataset name (leafname for storing old results)
# @param[in] $params        Parameters hashref for RSS generation:
#                           'title' => Page title
#                           'link'  => Link to related content
#                           'lang'  => Language of content
#                           'desc'  => Page description
#                           'logo'  => Page logo
#                           'logox' => Page logo width
#                           'logoy' => Page logo height
#                           'encoding' => Encoding of the content
#                           'format' => Format of change descriptions:
#                                       'pre'    : preformatted (<pre>)
#                                       'escape' : escapable text
#                                       'html'   : content is HTML
#                           'pre'   => Change descriptions are preformatted
sub generate
{
    my ($changes, $name, $params) = @_;
    
    my $rss = "";
    
    $rname = $name || 'random-stuff';
    $params = {} if (!defined $params);
    $params->{'title'} ||= "Changes for $rname";
    $params->{'link'}  ||= "x-no-url:$rname";
    $params->{'lang'}  ||= "en";
    $params->{'desc'}  ||= "Automagically generated list of changes of $rname";
    $params->{'logo'}  ||= "data:none";
    $params->{'logox'}  ||= "";
    $params->{'logoy'}  ||= "";
    $params->{'encoding'} ||= 'ISO-8859-1';
    $params->{'format'} ||= 'escape';

    $rss .= <<EOM;
<?xml version="1.0" encoding="$params->{'encoding'}" ?>
<rss xmlns:dc="http://purl.org/dc/elements/1.1/" version="0.91">

  <channel>
    <title>$params->{'title'}</title>
    <link>$params->{'link'}</link>
    <description>$params->{'desc'}</description>
    <language>$params->{'lang'}</language>

    <image>
      <title>$params->{'title'}</title>
      <url>$params->{'logo'}</url>
      <link>$params->{'link'}</link>
      <width>$params->{'logox'}</width>
      <height>$params->{'logoy'}</height>
    </image>
EOM

    my $store;
    my $allChanges = [];
    if (defined $name)
    {
        $store = "$Changes::storeDir/rss-$name";
        if (-f $store)
        {
            $allChanges = retrieve $store;
        }
    }

    # Ensure there's a date on the new changes
    for my $change (@$changes)
    {
        if (!defined $change->{'date'})
        {
            $change->{'date'} = convertTimeToISO8601(time());
        }
    }

    # Put the new changes at the front of the list
    unshift @$allChanges, @$changes;

    # Generate items in the RSS feed
    for my $change (@$allChanges)
    {
        my $title = $change->{'title'};
        my $desc = $change->{'desc'};
        my $date = $change->{'date'};

        # Make into HTML.
        if ($params->{'format'} eq 'pre' ||
            $params->{'format'} eq 'escape')
        {
            $desc =~ s/&/&amp;/g;
            $desc =~ s/</&lt;/g;
            $desc =~ s/>/&gt;/g;
        }
        if ($params->{'format'} eq 'escape')
        {
            $desc =~ s/^( +)/"&#160;" x length $1/gem;
            $desc =~ s/\n/<br>/g;
        }
        elsif ($params->{'format'} eq 'pre')
        {
            $desc = "<pre>$desc</pre>";
        }

        if ($type == 0)
        {
            # double escape - seems less reliable.
            $desc =~ s/</&lt;/g;
            $desc =~ s/>/&gt;/g;
            $desc =~ s/&/&amp;/g;
        }
        else
        {
            $desc = "<![CDATA[$desc]]>";
        }

        my $md5 = Digest::MD5::md5_hex($title.$desc.$date);

        $rss .= <<EOM;
    <item>
      <title>$date - $title</title>
      <link>x-notvalid:$rname#$md5</link>
      <description>$desc</description>
      <dc:date>$date</dc:date>
    </item>
EOM
    }

    $rss .= <<EOM;

  </channel>
</rss>
EOM

    if (defined $name)
    {
        store $allChanges, $store;
    }
    
    return $rss;
}

##
# Convert the time supplied in UTC unix time to an ISO 8601
# format time string.
#
# @param[in]  $now   the time to convert
#
# @return the string for this time
sub convertTimeToISO8601
{
    my ($now) = @_;
    my ($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst) =
        gmtime($now);
    $year += 1900;
    $mon  += 1;
    return sprintf "%4d-%02d-%02dT%02d:%02dZ",
                   $year, $mon, $mday, $hour, $min;
}
