#!/usr/bin/perl -w
#
# Get details from passwd into a hashref.
#
# Example use:
#
# my $users = Passwd::readActive();
#
# for my $username (sort keys %$users)
# {
#   my $user = $users->{$user};
#   print "User: $username    Shell: $user->{'shell'}\n";
# }
#

package Passwd;

# For debugging purposes:
if ($0 =~ /Passwd\.pm/)
{
    $users = Passwd::read();
    $users = Passwd::readActive();
    print map { "$_ $users->{$_}->{'shell'}\n" } keys %$users;
}


##
# Read the password file and return a hashref containing
# the entries.
# 
# @return Hashref, keyed by user name containing:
#           'username'  => the user name
#           'uid'       => uid number
#           'gid'       => gid number
#           'name'      => user name
#           'home'      => home directory
#           'shell'     => shell to use
#           'room'      => room number/name
#           'home#'     => Home phone number
#           'work#'     => Work phone number
sub read
{
    my %user;
    my $passwd = `getent passwd`;
    for my $line (split /\n/, $passwd)
    {
        my ($username, $pass, $uid, $gid, $nameent, $home, $shell) = split /:/, $line;
        next if ($username =~ /\$/); # skip machines
        next if ($shell eq ""); # skip users without shells
        my ($name, $room, $worknum, $homenum) = split /,/, $nameent;
        $user{$username} = {
            'username' => $username,
            'uid'      => $uid,
            'gid'      => $gid,
            'name'     => $name,
            'room'     => $room,
            'home#'    => $homenum,
            'work#'    => $worknum,
            'home'     => $home,
            'shell'    => $shell
        };
    }
    if ((keys %user) == 0)
    {
        die "No users found; giving up\n";
    }

    return \%user;
}


##
# Read the password file, skipping out users which have been
# flagged as disabled or as having left (Picsel style; duplicated
# here for simplicity).
#
# @return Hashref, as per read().
sub readActive
{
    my $users = Passwd::read();
    return { map { $_ => $users->{$_} }
                grep { $users->{$_}->{'shell'} !~ /false/ &&
                       $users->{$_}->{'name'} !~ /departed/i }
                    keys %$users };
}

##
# Describe a password entry.
#
# @param[in] $user  Hashref of the entry to describe
sub describe
{
    my ($user) = @_;
    return "Username: $user->{'username'}\n" .
           "Name:     $user->{'name'}\n" .
           "ID:       $user->{'uid'} / $user->{'gid'}\n" .
           "Shell:    $user->{'shell'}\n" .
           "Home:     $user->{'home'}\n";
}

1;
